/* File:      gpp.h  -- generic preprocessor
** Author:    Denis Auroux, Tristan Miller, Joel Bertrand
** Contact:   psychonaut@nothingisreal.com
** 
** Copyright (C) 1996, 1999, 2001 Denis Auroux
** Copyright (C) 2003, 2004 Tristan Miller
** Copyright (C) 2011 Joel Bertrand
** 
** GPP is free software; you can redistribute it and/or modify it under the
** terms of the GNU Lesser General Public License as published by the Free
** Software Foundation; either version 2.1 of the License, or (at your option)
** any later version.
** 
** GPP is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
** more details.
** 
** You should have received a copy of the GNU Lesser General Public License
** along with this software; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: gpp.c,v 1.10 2004/09/19 20:19:16 psy Exp $
** 
*/

/* To compile under MS VC++, one must define WIN_NT */

#ifndef _GPP_H_
#	define _GPP_H_

#	if HAVE_CONFIG_H
#		include <config.h>
#	endif

#	ifdef WIN_NT				/* WIN NT settings */
#		define popen			_popen
#		define pclose			_pclose
#		define my_strdup		_strdup
#		define my_strcasecmp	_stricmp
#		define SLASH			'\\'
#		define DEFAULT_CRLF		1
#	else						/* UNIX settings */
#		define SLASH			'/'
#		define DEFAULT_CRLF		0
#	endif

#	include <stdio.h>
#	include <stdlib.h>
#	include <string.h>
#	include <ctype.h>
#	if HAVE_FNMATCH_H
#		include <fnmatch.h>
#	endif
#	include <time.h>

#	define STACKDEPTH				50
#	define MAXARGS					100
#	define MAXINCL					64	/* max # of include dirs */

#	define MAX_GPP_NUM_SIZE		15
#	define MAX_GPP_DATE_SIZE		1024

	typedef struct MODE
	{
		char *mStart;				/* before macro name */
		char *mEnd;					/* end macro without arg */
		char *mArgS;				/* start 1st argument */
		char *mArgSep;				/* separate arguments */
		char *mArgE;				/* end last argument */
		char *mArgRef;				/* how to refer to arguments in a def */
		char quotechar;				/* quote next char */
		char *stackchar;			/* characters to stack */
		char *unstackchar;			/* characters to unstack */
	} MODE;

/*
   translation for delimiters :
   \001 = \b = ' ' = one or more spaces    \201 = \!b = non-space
   \002 = \w = zero or more spaces 
   \003 = \B = one or more spaces or \n    \203 = \!B = non-space nor \n
   \004 = \W = zero or more spaces or \n 
   \005 = \a = alphabetic (a-z, A-Z)       \205 = \!a = non-alphabetic
   \006 = \A = alphabetic or space/\n      \206 = \!A
   \007 = \# = numeric (0-9)               \207 = \!#
   \010 = \i = identifier (a-zA-Z0-9_)     \210 = \!i
   \011 = \t, \012 = \n                    \211 = \!t, \212 = \!n
   \013 = \o = operator (+-*\/^<>=`~:.?@#&!%|) \213 = \!o
   \014 = \O = operator or ()[]{}              \214 = \!O
*/

/*        st        end   args   sep    arge ref  quot  stk  unstk*/
	struct MODE CUser =
		{ "",       "",   "(",   ",",   ")", "#", '\\', "(", ")" };
	struct MODE CMeta =
		{ "#",      "\n", "\001","\001","\n","#", '\\', "(", ")" };
	struct MODE KUser =
		{ "",       "",   "(",   ",",   ")", "#",  0,   "(", ")" };
	struct MODE KMeta =
		{ "\n#\002","\n", "\001","\001","\n","#",  0,   "",  ""  }; 
	struct MODE Tex   =
		{ "\\",     "",   "{",   "}{",  "}", "#", '@',  "{", "}" };
	struct MODE Html  =
		{ "<#",     ">",  "\003","|",   ">", "#", '\\', "<", ">" };
	struct MODE XHtml =
		{ "<#",     "/>", "\003","|",   "/>","#", '\\', "<", ">" };

#	define DEFAULT_OP_STRING	((unsigned char *) "+-*/\\^<>=`~:.?@#&!%|")
#	define PROLOG_OP_STRING		((unsigned char *) "+-*/\\^<>=`~:.?@#&")
#	define DEFAULT_OP_PLUS		((unsigned char *) "()[]{}")
#	define DEFAULT_ID_STRING	((unsigned char *) "\005\007_")
/* or equiv. "A-Za-z0-9_" */

/* here we assume that longs are at least 32 bit... if not, change this ! */
#	define LOG_LONG_BITS			5
#	define CHARSET_SUBSET_LEN		(256 >> LOG_LONG_BITS)
	typedef unsigned long *CHARSET_SUBSET;

	CHARSET_SUBSET DefaultOp, DefaultExtOp, PrologOp, DefaultId;

	typedef struct COMMENT
	{
		char *start;			/* how the comment/string starts */
		char *end;			/* how it ends */
		char quote;			/* how to prevent it from ending */
		char warn;			/* a character that shouldn't be in there */
		int flags[3];			/* meta, user, text */
		struct COMMENT *next;
	} COMMENT;

#	define OUTPUT_TEXT		0x1   /* what's inside will be output */
#	define OUTPUT_DELIM	0x2   /* the delimiters will be output */
#	define PARSE_MACROS	0x4   /* macros inside will be parsed */
#	define FLAG_IGNORE		0x40 

#	define FLAG_STRING		(OUTPUT_TEXT|OUTPUT_DELIM)
#	define FLAG_COMMENT	0

#	define FLAG_META		0
#	define FLAG_USER		1
#	define FLAG_TEXT		2

/* Some stuff I removed because it made for some impossible situations :

 #define PARSE_COMMENTS  0x8   
   comments inside comments will not be parsed because nesting comments is 
   too complicated (syntax conflicts, esp. to find a comment's end)
   -- of course, unless the comment is ignored.
   
 #define MACRO_FRIENDLY  0x20  
   a comment-end is to be processed even if an unfinished macro call has 
   started inside the comment, otherwise it's too hard do decide in advance 
   where a comment ends. In particular foo('bar((((') is valid.

 #define PREVENT_DELIM   0x10 
   all comments will prevent macro delimitation, i.e. foo('bar) is invalid.
   -- of course, unless the comment is ignored.
   Too bad, #define foo '...    terminates only at following "'".
   Unless one adds quotechars like in #define foo \' ...
   
 ALSO NOTE : comments are not allowed before the end of the first argument
 to a meta-macro. E.g. this is legal :   #define foo <* blah *> 3
 This is not legal :                     #define <* blah *> foo 3
 If a comment occurs here, the behavior depends on the actual meta-macro :
 most will yield an error and stop gpp (#define, #undef, #ifdef/ifndef, 
 #defeval, #include, #mode) ; #exec, #if and #eval should be ok ; 
 #ifeq will always fail while #ifneq will always succeed ;
*/ 

	typedef struct SPECS
	{
		struct MODE		User;
		struct MODE		Meta;
		struct COMMENT	*comments;
		struct SPECS	*stack_next;
		int				preservelf;
		CHARSET_SUBSET	op_set,ext_op_set,id_set;
	} SPECS;

	struct SPECS		*S;

	typedef struct MACRO
	{
		char			*username;
		char			*macrotext;
		char			**argnames;
		int				macrolen;
		int				nnamedargs;
		struct SPECS	*define_specs;
		int				defined_in_comment;
		int				protect;
	} MACRO;

	struct MACRO		*macros;
	int					nmacros;
	int					nalloced;
	char				*includedir[MAXINCL];
	int					nincludedirs;
	int					execallowed;
	int					dosmode;
	int					autoswitch;
/* must be a format-like string that has % % % in it.
   The first % is replaced with line number, the second with "filename", and
   the third with 1, 2 or blank
   Can also use ? instead of %.
*/
	char				*include_directive_marker = NULL;
	short				WarningLevel = 2;

/* controls if standard dirs, like /usr/include, are to be searched for
   #include and whether the current dir is to be searched first or last. */
	int					NoStdInc        = 0;
	int					NoCurIncFirst   = 0;
	int					CurDirIncLast   = 0;
	int					file_and_stdout = 0;
	char				*IncludeFile   = NULL;

	typedef struct OUTPUTCONTEXT
	{
		char *buf;
		int len,bufsize;
		FILE *f;
	} OUTPUTCONTEXT;

	typedef struct INPUTCONTEXT
	{
		char					*buf;
		char					*malloced_buf;
				/* what was actually malloc-ed (buf may have shifted) */
		int						len,bufsize;
		int						lineno;
		char					*filename;
		FILE					*in;
		int						argc;
		char					**argv;
		char					**namedargs;
		struct OUTPUTCONTEXT	 *out;
		int						eof;
		int						in_comment;
		int						ambience;
				/* FLAG_TEXT, FLAG_USER or FLAG_META */
		int						may_have_args;
	} INPUTCONTEXT;

	struct INPUTCONTEXT			*C;
  
	int							commented[STACKDEPTH];
	int							iflevel;

/* commented = 0: output, 1: not output, 
   2: not output because we're in a #elif and we've already gone through
      the right case (so #else/#elif can't toggle back to output) */

	void ProcessContext(void); /* the main loop */

	int findIdent(const char *b,int l);
	void delete_macro(int i);

/* various recent additions */
	void usage(void);
	void display_version(void);
	void bug(const char *s);
	void warning(const char *s);
	static void getDirname(const char *fname, char *dirname);
	static FILE *openInCurrentDir(const char *incfile);
	char *ArithmEval(int pos1,int pos2);
	void replace_definition_with_blank_lines(const char *start,
			const char *end, int skip);
	void replace_directive_with_blank_line(FILE *file);
	void write_include_marker(FILE *f, int lineno, char *filename,
			const char *marker);
	void construct_include_directive_marker(char **include_directive_marker,
			const char *includemarker_input);
	void escape_backslashes(const char *instr, char **outstr);
	static void DoInclude(char *file_name);

/*
** strdup() and my_strcasecmp() are not ANSI C, so here we define our own
** versions in case the compiler does not support them
*/
#	if ! HAVE_STRDUP
	inline char *my_strdup(const char *s)
	{
		size_t len = strlen(s) + 1;
		char *newstr = malloc(len);
		return newstr ? (char *)memcpy(newstr, s, len) : NULL;
	}
#	else
#		undef my_strdup
#		define my_strdup strdup
#	endif

#	if ! HAVE_STRCASECMP
	inline int my_strcasecmp(const char *s, const char *s2)
	{
		do
		{
			char c1 = tolower(*s);
			char c2 = tolower(*s2);

			if (c1 > c2) return 1;
			if (c1 < c2) return -1;
		} while(*s++ && *s2++);

		return 0;
		}
#	else
#		undef my_strcasecmp
#		define my_strcasecmp strcasecmp
#	endif
#endif

// vim: ts=4
